"""
polish_scores

====

This module provides a way to download and process the Polish Scores dataset.
"""

from .download_polish_dataset import (
    convert_and_download_file,
    main as download_polish_scores,
)

