"""
# kernpy

=====


Python Humdrum **kern and **mens utilities package.



Execute the following command to run **kernpy** as a module:
```shell
python -m kernpy --help
python -m kernpy <command> <options>
```
"""


from .core import *

from .io import *

from .util import *

from .polish_scores import *





