# Generated from kernSpineParser.g4 by ANTLR 4.13.1
from antlr4 import *
if "." in __name__:
    from .kernSpineParser import kernSpineParser
else:
    from kernSpineParser import kernSpineParser

# This class defines a complete listener for a parse tree produced by kernSpineParser.
class kernSpineParserListener(ParseTreeListener):

    # Enter a parse tree produced by kernSpineParser#start.
    def enterStart(self, ctx:kernSpineParser.StartContext):
        pass

    # Exit a parse tree produced by kernSpineParser#start.
    def exitStart(self, ctx:kernSpineParser.StartContext):
        pass


    # Enter a parse tree produced by kernSpineParser#field.
    def enterField(self, ctx:kernSpineParser.FieldContext):
        pass

    # Exit a parse tree produced by kernSpineParser#field.
    def exitField(self, ctx:kernSpineParser.FieldContext):
        pass


    # Enter a parse tree produced by kernSpineParser#notes_rests_chords.
    def enterNotes_rests_chords(self, ctx:kernSpineParser.Notes_rests_chordsContext):
        pass

    # Exit a parse tree produced by kernSpineParser#notes_rests_chords.
    def exitNotes_rests_chords(self, ctx:kernSpineParser.Notes_rests_chordsContext):
        pass


    # Enter a parse tree produced by kernSpineParser#structural.
    def enterStructural(self, ctx:kernSpineParser.StructuralContext):
        pass

    # Exit a parse tree produced by kernSpineParser#structural.
    def exitStructural(self, ctx:kernSpineParser.StructuralContext):
        pass


    # Enter a parse tree produced by kernSpineParser#contextual.
    def enterContextual(self, ctx:kernSpineParser.ContextualContext):
        pass

    # Exit a parse tree produced by kernSpineParser#contextual.
    def exitContextual(self, ctx:kernSpineParser.ContextualContext):
        pass


    # Enter a parse tree produced by kernSpineParser#signatures.
    def enterSignatures(self, ctx:kernSpineParser.SignaturesContext):
        pass

    # Exit a parse tree produced by kernSpineParser#signatures.
    def exitSignatures(self, ctx:kernSpineParser.SignaturesContext):
        pass


    # Enter a parse tree produced by kernSpineParser#otherContextual.
    def enterOtherContextual(self, ctx:kernSpineParser.OtherContextualContext):
        pass

    # Exit a parse tree produced by kernSpineParser#otherContextual.
    def exitOtherContextual(self, ctx:kernSpineParser.OtherContextualContext):
        pass


    # Enter a parse tree produced by kernSpineParser#empty.
    def enterEmpty(self, ctx:kernSpineParser.EmptyContext):
        pass

    # Exit a parse tree produced by kernSpineParser#empty.
    def exitEmpty(self, ctx:kernSpineParser.EmptyContext):
        pass


    # Enter a parse tree produced by kernSpineParser#rest.
    def enterRest(self, ctx:kernSpineParser.RestContext):
        pass

    # Exit a parse tree produced by kernSpineParser#rest.
    def exitRest(self, ctx:kernSpineParser.RestContext):
        pass


    # Enter a parse tree produced by kernSpineParser#restChar_r.
    def enterRestChar_r(self, ctx:kernSpineParser.RestChar_rContext):
        pass

    # Exit a parse tree produced by kernSpineParser#restChar_r.
    def exitRestChar_r(self, ctx:kernSpineParser.RestChar_rContext):
        pass


    # Enter a parse tree produced by kernSpineParser#restDecoration.
    def enterRestDecoration(self, ctx:kernSpineParser.RestDecorationContext):
        pass

    # Exit a parse tree produced by kernSpineParser#restDecoration.
    def exitRestDecoration(self, ctx:kernSpineParser.RestDecorationContext):
        pass


    # Enter a parse tree produced by kernSpineParser#chord.
    def enterChord(self, ctx:kernSpineParser.ChordContext):
        pass

    # Exit a parse tree produced by kernSpineParser#chord.
    def exitChord(self, ctx:kernSpineParser.ChordContext):
        pass


    # Enter a parse tree produced by kernSpineParser#note.
    def enterNote(self, ctx:kernSpineParser.NoteContext):
        pass

    # Exit a parse tree produced by kernSpineParser#note.
    def exitNote(self, ctx:kernSpineParser.NoteContext):
        pass


    # Enter a parse tree produced by kernSpineParser#nonVisualTandemInterpretation.
    def enterNonVisualTandemInterpretation(self, ctx:kernSpineParser.NonVisualTandemInterpretationContext):
        pass

    # Exit a parse tree produced by kernSpineParser#nonVisualTandemInterpretation.
    def exitNonVisualTandemInterpretation(self, ctx:kernSpineParser.NonVisualTandemInterpretationContext):
        pass


    # Enter a parse tree produced by kernSpineParser#boundingBox.
    def enterBoundingBox(self, ctx:kernSpineParser.BoundingBoxContext):
        pass

    # Exit a parse tree produced by kernSpineParser#boundingBox.
    def exitBoundingBox(self, ctx:kernSpineParser.BoundingBoxContext):
        pass


    # Enter a parse tree produced by kernSpineParser#xywh.
    def enterXywh(self, ctx:kernSpineParser.XywhContext):
        pass

    # Exit a parse tree produced by kernSpineParser#xywh.
    def exitXywh(self, ctx:kernSpineParser.XywhContext):
        pass


    # Enter a parse tree produced by kernSpineParser#x.
    def enterX(self, ctx:kernSpineParser.XContext):
        pass

    # Exit a parse tree produced by kernSpineParser#x.
    def exitX(self, ctx:kernSpineParser.XContext):
        pass


    # Enter a parse tree produced by kernSpineParser#y.
    def enterY(self, ctx:kernSpineParser.YContext):
        pass

    # Exit a parse tree produced by kernSpineParser#y.
    def exitY(self, ctx:kernSpineParser.YContext):
        pass


    # Enter a parse tree produced by kernSpineParser#w.
    def enterW(self, ctx:kernSpineParser.WContext):
        pass

    # Exit a parse tree produced by kernSpineParser#w.
    def exitW(self, ctx:kernSpineParser.WContext):
        pass


    # Enter a parse tree produced by kernSpineParser#h.
    def enterH(self, ctx:kernSpineParser.HContext):
        pass

    # Exit a parse tree produced by kernSpineParser#h.
    def exitH(self, ctx:kernSpineParser.HContext):
        pass


    # Enter a parse tree produced by kernSpineParser#pageNumber.
    def enterPageNumber(self, ctx:kernSpineParser.PageNumberContext):
        pass

    # Exit a parse tree produced by kernSpineParser#pageNumber.
    def exitPageNumber(self, ctx:kernSpineParser.PageNumberContext):
        pass


    # Enter a parse tree produced by kernSpineParser#visualTandemInterpretation.
    def enterVisualTandemInterpretation(self, ctx:kernSpineParser.VisualTandemInterpretationContext):
        pass

    # Exit a parse tree produced by kernSpineParser#visualTandemInterpretation.
    def exitVisualTandemInterpretation(self, ctx:kernSpineParser.VisualTandemInterpretationContext):
        pass


    # Enter a parse tree produced by kernSpineParser#associatedIDS.
    def enterAssociatedIDS(self, ctx:kernSpineParser.AssociatedIDSContext):
        pass

    # Exit a parse tree produced by kernSpineParser#associatedIDS.
    def exitAssociatedIDS(self, ctx:kernSpineParser.AssociatedIDSContext):
        pass


    # Enter a parse tree produced by kernSpineParser#placeHolder.
    def enterPlaceHolder(self, ctx:kernSpineParser.PlaceHolderContext):
        pass

    # Exit a parse tree produced by kernSpineParser#placeHolder.
    def exitPlaceHolder(self, ctx:kernSpineParser.PlaceHolderContext):
        pass


    # Enter a parse tree produced by kernSpineParser#octaveShift.
    def enterOctaveShift(self, ctx:kernSpineParser.OctaveShiftContext):
        pass

    # Exit a parse tree produced by kernSpineParser#octaveShift.
    def exitOctaveShift(self, ctx:kernSpineParser.OctaveShiftContext):
        pass


    # Enter a parse tree produced by kernSpineParser#pianoHand.
    def enterPianoHand(self, ctx:kernSpineParser.PianoHandContext):
        pass

    # Exit a parse tree produced by kernSpineParser#pianoHand.
    def exitPianoHand(self, ctx:kernSpineParser.PianoHandContext):
        pass


    # Enter a parse tree produced by kernSpineParser#tandemTuplet.
    def enterTandemTuplet(self, ctx:kernSpineParser.TandemTupletContext):
        pass

    # Exit a parse tree produced by kernSpineParser#tandemTuplet.
    def exitTandemTuplet(self, ctx:kernSpineParser.TandemTupletContext):
        pass


    # Enter a parse tree produced by kernSpineParser#tandemCue.
    def enterTandemCue(self, ctx:kernSpineParser.TandemCueContext):
        pass

    # Exit a parse tree produced by kernSpineParser#tandemCue.
    def exitTandemCue(self, ctx:kernSpineParser.TandemCueContext):
        pass


    # Enter a parse tree produced by kernSpineParser#tandemTremolo.
    def enterTandemTremolo(self, ctx:kernSpineParser.TandemTremoloContext):
        pass

    # Exit a parse tree produced by kernSpineParser#tandemTremolo.
    def exitTandemTremolo(self, ctx:kernSpineParser.TandemTremoloContext):
        pass


    # Enter a parse tree produced by kernSpineParser#ossia.
    def enterOssia(self, ctx:kernSpineParser.OssiaContext):
        pass

    # Exit a parse tree produced by kernSpineParser#ossia.
    def exitOssia(self, ctx:kernSpineParser.OssiaContext):
        pass


    # Enter a parse tree produced by kernSpineParser#rscale.
    def enterRscale(self, ctx:kernSpineParser.RscaleContext):
        pass

    # Exit a parse tree produced by kernSpineParser#rscale.
    def exitRscale(self, ctx:kernSpineParser.RscaleContext):
        pass


    # Enter a parse tree produced by kernSpineParser#pedal.
    def enterPedal(self, ctx:kernSpineParser.PedalContext):
        pass

    # Exit a parse tree produced by kernSpineParser#pedal.
    def exitPedal(self, ctx:kernSpineParser.PedalContext):
        pass


    # Enter a parse tree produced by kernSpineParser#ela.
    def enterEla(self, ctx:kernSpineParser.ElaContext):
        pass

    # Exit a parse tree produced by kernSpineParser#ela.
    def exitEla(self, ctx:kernSpineParser.ElaContext):
        pass


    # Enter a parse tree produced by kernSpineParser#dynamics_position.
    def enterDynamics_position(self, ctx:kernSpineParser.Dynamics_positionContext):
        pass

    # Exit a parse tree produced by kernSpineParser#dynamics_position.
    def exitDynamics_position(self, ctx:kernSpineParser.Dynamics_positionContext):
        pass


    # Enter a parse tree produced by kernSpineParser#sections.
    def enterSections(self, ctx:kernSpineParser.SectionsContext):
        pass

    # Exit a parse tree produced by kernSpineParser#sections.
    def exitSections(self, ctx:kernSpineParser.SectionsContext):
        pass


    # Enter a parse tree produced by kernSpineParser#sectionNames.
    def enterSectionNames(self, ctx:kernSpineParser.SectionNamesContext):
        pass

    # Exit a parse tree produced by kernSpineParser#sectionNames.
    def exitSectionNames(self, ctx:kernSpineParser.SectionNamesContext):
        pass


    # Enter a parse tree produced by kernSpineParser#sectionName.
    def enterSectionName(self, ctx:kernSpineParser.SectionNameContext):
        pass

    # Exit a parse tree produced by kernSpineParser#sectionName.
    def exitSectionName(self, ctx:kernSpineParser.SectionNameContext):
        pass


    # Enter a parse tree produced by kernSpineParser#transposition.
    def enterTransposition(self, ctx:kernSpineParser.TranspositionContext):
        pass

    # Exit a parse tree produced by kernSpineParser#transposition.
    def exitTransposition(self, ctx:kernSpineParser.TranspositionContext):
        pass


    # Enter a parse tree produced by kernSpineParser#instrument.
    def enterInstrument(self, ctx:kernSpineParser.InstrumentContext):
        pass

    # Exit a parse tree produced by kernSpineParser#instrument.
    def exitInstrument(self, ctx:kernSpineParser.InstrumentContext):
        pass


    # Enter a parse tree produced by kernSpineParser#instrumentTitle.
    def enterInstrumentTitle(self, ctx:kernSpineParser.InstrumentTitleContext):
        pass

    # Exit a parse tree produced by kernSpineParser#instrumentTitle.
    def exitInstrumentTitle(self, ctx:kernSpineParser.InstrumentTitleContext):
        pass


    # Enter a parse tree produced by kernSpineParser#number.
    def enterNumber(self, ctx:kernSpineParser.NumberContext):
        pass

    # Exit a parse tree produced by kernSpineParser#number.
    def exitNumber(self, ctx:kernSpineParser.NumberContext):
        pass


    # Enter a parse tree produced by kernSpineParser#lowerCasePitch.
    def enterLowerCasePitch(self, ctx:kernSpineParser.LowerCasePitchContext):
        pass

    # Exit a parse tree produced by kernSpineParser#lowerCasePitch.
    def exitLowerCasePitch(self, ctx:kernSpineParser.LowerCasePitchContext):
        pass


    # Enter a parse tree produced by kernSpineParser#upperCasePitch.
    def enterUpperCasePitch(self, ctx:kernSpineParser.UpperCasePitchContext):
        pass

    # Exit a parse tree produced by kernSpineParser#upperCasePitch.
    def exitUpperCasePitch(self, ctx:kernSpineParser.UpperCasePitchContext):
        pass


    # Enter a parse tree produced by kernSpineParser#pitchClass.
    def enterPitchClass(self, ctx:kernSpineParser.PitchClassContext):
        pass

    # Exit a parse tree produced by kernSpineParser#pitchClass.
    def exitPitchClass(self, ctx:kernSpineParser.PitchClassContext):
        pass


    # Enter a parse tree produced by kernSpineParser#accomp.
    def enterAccomp(self, ctx:kernSpineParser.AccompContext):
        pass

    # Exit a parse tree produced by kernSpineParser#accomp.
    def exitAccomp(self, ctx:kernSpineParser.AccompContext):
        pass


    # Enter a parse tree produced by kernSpineParser#solo.
    def enterSolo(self, ctx:kernSpineParser.SoloContext):
        pass

    # Exit a parse tree produced by kernSpineParser#solo.
    def exitSolo(self, ctx:kernSpineParser.SoloContext):
        pass


    # Enter a parse tree produced by kernSpineParser#strophe.
    def enterStrophe(self, ctx:kernSpineParser.StropheContext):
        pass

    # Exit a parse tree produced by kernSpineParser#strophe.
    def exitStrophe(self, ctx:kernSpineParser.StropheContext):
        pass


    # Enter a parse tree produced by kernSpineParser#timebase.
    def enterTimebase(self, ctx:kernSpineParser.TimebaseContext):
        pass

    # Exit a parse tree produced by kernSpineParser#timebase.
    def exitTimebase(self, ctx:kernSpineParser.TimebaseContext):
        pass


    # Enter a parse tree produced by kernSpineParser#part.
    def enterPart(self, ctx:kernSpineParser.PartContext):
        pass

    # Exit a parse tree produced by kernSpineParser#part.
    def exitPart(self, ctx:kernSpineParser.PartContext):
        pass


    # Enter a parse tree produced by kernSpineParser#group.
    def enterGroup(self, ctx:kernSpineParser.GroupContext):
        pass

    # Exit a parse tree produced by kernSpineParser#group.
    def exitGroup(self, ctx:kernSpineParser.GroupContext):
        pass


    # Enter a parse tree produced by kernSpineParser#staff.
    def enterStaff(self, ctx:kernSpineParser.StaffContext):
        pass

    # Exit a parse tree produced by kernSpineParser#staff.
    def exitStaff(self, ctx:kernSpineParser.StaffContext):
        pass


    # Enter a parse tree produced by kernSpineParser#clef.
    def enterClef(self, ctx:kernSpineParser.ClefContext):
        pass

    # Exit a parse tree produced by kernSpineParser#clef.
    def exitClef(self, ctx:kernSpineParser.ClefContext):
        pass


    # Enter a parse tree produced by kernSpineParser#clefValue.
    def enterClefValue(self, ctx:kernSpineParser.ClefValueContext):
        pass

    # Exit a parse tree produced by kernSpineParser#clefValue.
    def exitClefValue(self, ctx:kernSpineParser.ClefValueContext):
        pass


    # Enter a parse tree produced by kernSpineParser#clefSign.
    def enterClefSign(self, ctx:kernSpineParser.ClefSignContext):
        pass

    # Exit a parse tree produced by kernSpineParser#clefSign.
    def exitClefSign(self, ctx:kernSpineParser.ClefSignContext):
        pass


    # Enter a parse tree produced by kernSpineParser#clefLine.
    def enterClefLine(self, ctx:kernSpineParser.ClefLineContext):
        pass

    # Exit a parse tree produced by kernSpineParser#clefLine.
    def exitClefLine(self, ctx:kernSpineParser.ClefLineContext):
        pass


    # Enter a parse tree produced by kernSpineParser#clefOctave.
    def enterClefOctave(self, ctx:kernSpineParser.ClefOctaveContext):
        pass

    # Exit a parse tree produced by kernSpineParser#clefOctave.
    def exitClefOctave(self, ctx:kernSpineParser.ClefOctaveContext):
        pass


    # Enter a parse tree produced by kernSpineParser#keySignature.
    def enterKeySignature(self, ctx:kernSpineParser.KeySignatureContext):
        pass

    # Exit a parse tree produced by kernSpineParser#keySignature.
    def exitKeySignature(self, ctx:kernSpineParser.KeySignatureContext):
        pass


    # Enter a parse tree produced by kernSpineParser#keySignaturePitchClass.
    def enterKeySignaturePitchClass(self, ctx:kernSpineParser.KeySignaturePitchClassContext):
        pass

    # Exit a parse tree produced by kernSpineParser#keySignaturePitchClass.
    def exitKeySignaturePitchClass(self, ctx:kernSpineParser.KeySignaturePitchClassContext):
        pass


    # Enter a parse tree produced by kernSpineParser#keySignatureCancel.
    def enterKeySignatureCancel(self, ctx:kernSpineParser.KeySignatureCancelContext):
        pass

    # Exit a parse tree produced by kernSpineParser#keySignatureCancel.
    def exitKeySignatureCancel(self, ctx:kernSpineParser.KeySignatureCancelContext):
        pass


    # Enter a parse tree produced by kernSpineParser#keyCancel.
    def enterKeyCancel(self, ctx:kernSpineParser.KeyCancelContext):
        pass

    # Exit a parse tree produced by kernSpineParser#keyCancel.
    def exitKeyCancel(self, ctx:kernSpineParser.KeyCancelContext):
        pass


    # Enter a parse tree produced by kernSpineParser#keyMode.
    def enterKeyMode(self, ctx:kernSpineParser.KeyModeContext):
        pass

    # Exit a parse tree produced by kernSpineParser#keyMode.
    def exitKeyMode(self, ctx:kernSpineParser.KeyModeContext):
        pass


    # Enter a parse tree produced by kernSpineParser#key.
    def enterKey(self, ctx:kernSpineParser.KeyContext):
        pass

    # Exit a parse tree produced by kernSpineParser#key.
    def exitKey(self, ctx:kernSpineParser.KeyContext):
        pass


    # Enter a parse tree produced by kernSpineParser#singleKey.
    def enterSingleKey(self, ctx:kernSpineParser.SingleKeyContext):
        pass

    # Exit a parse tree produced by kernSpineParser#singleKey.
    def exitSingleKey(self, ctx:kernSpineParser.SingleKeyContext):
        pass


    # Enter a parse tree produced by kernSpineParser#minorKey.
    def enterMinorKey(self, ctx:kernSpineParser.MinorKeyContext):
        pass

    # Exit a parse tree produced by kernSpineParser#minorKey.
    def exitMinorKey(self, ctx:kernSpineParser.MinorKeyContext):
        pass


    # Enter a parse tree produced by kernSpineParser#majorKey.
    def enterMajorKey(self, ctx:kernSpineParser.MajorKeyContext):
        pass

    # Exit a parse tree produced by kernSpineParser#majorKey.
    def exitMajorKey(self, ctx:kernSpineParser.MajorKeyContext):
        pass


    # Enter a parse tree produced by kernSpineParser#modal.
    def enterModal(self, ctx:kernSpineParser.ModalContext):
        pass

    # Exit a parse tree produced by kernSpineParser#modal.
    def exitModal(self, ctx:kernSpineParser.ModalContext):
        pass


    # Enter a parse tree produced by kernSpineParser#locrian.
    def enterLocrian(self, ctx:kernSpineParser.LocrianContext):
        pass

    # Exit a parse tree produced by kernSpineParser#locrian.
    def exitLocrian(self, ctx:kernSpineParser.LocrianContext):
        pass


    # Enter a parse tree produced by kernSpineParser#ionian.
    def enterIonian(self, ctx:kernSpineParser.IonianContext):
        pass

    # Exit a parse tree produced by kernSpineParser#ionian.
    def exitIonian(self, ctx:kernSpineParser.IonianContext):
        pass


    # Enter a parse tree produced by kernSpineParser#aeolian.
    def enterAeolian(self, ctx:kernSpineParser.AeolianContext):
        pass

    # Exit a parse tree produced by kernSpineParser#aeolian.
    def exitAeolian(self, ctx:kernSpineParser.AeolianContext):
        pass


    # Enter a parse tree produced by kernSpineParser#mixolydian.
    def enterMixolydian(self, ctx:kernSpineParser.MixolydianContext):
        pass

    # Exit a parse tree produced by kernSpineParser#mixolydian.
    def exitMixolydian(self, ctx:kernSpineParser.MixolydianContext):
        pass


    # Enter a parse tree produced by kernSpineParser#lydian.
    def enterLydian(self, ctx:kernSpineParser.LydianContext):
        pass

    # Exit a parse tree produced by kernSpineParser#lydian.
    def exitLydian(self, ctx:kernSpineParser.LydianContext):
        pass


    # Enter a parse tree produced by kernSpineParser#phrygian.
    def enterPhrygian(self, ctx:kernSpineParser.PhrygianContext):
        pass

    # Exit a parse tree produced by kernSpineParser#phrygian.
    def exitPhrygian(self, ctx:kernSpineParser.PhrygianContext):
        pass


    # Enter a parse tree produced by kernSpineParser#dorian.
    def enterDorian(self, ctx:kernSpineParser.DorianContext):
        pass

    # Exit a parse tree produced by kernSpineParser#dorian.
    def exitDorian(self, ctx:kernSpineParser.DorianContext):
        pass


    # Enter a parse tree produced by kernSpineParser#timeSignature.
    def enterTimeSignature(self, ctx:kernSpineParser.TimeSignatureContext):
        pass

    # Exit a parse tree produced by kernSpineParser#timeSignature.
    def exitTimeSignature(self, ctx:kernSpineParser.TimeSignatureContext):
        pass


    # Enter a parse tree produced by kernSpineParser#numerator.
    def enterNumerator(self, ctx:kernSpineParser.NumeratorContext):
        pass

    # Exit a parse tree produced by kernSpineParser#numerator.
    def exitNumerator(self, ctx:kernSpineParser.NumeratorContext):
        pass


    # Enter a parse tree produced by kernSpineParser#denominator.
    def enterDenominator(self, ctx:kernSpineParser.DenominatorContext):
        pass

    # Exit a parse tree produced by kernSpineParser#denominator.
    def exitDenominator(self, ctx:kernSpineParser.DenominatorContext):
        pass


    # Enter a parse tree produced by kernSpineParser#standardTimeSignature.
    def enterStandardTimeSignature(self, ctx:kernSpineParser.StandardTimeSignatureContext):
        pass

    # Exit a parse tree produced by kernSpineParser#standardTimeSignature.
    def exitStandardTimeSignature(self, ctx:kernSpineParser.StandardTimeSignatureContext):
        pass


    # Enter a parse tree produced by kernSpineParser#additiveTimeSignature.
    def enterAdditiveTimeSignature(self, ctx:kernSpineParser.AdditiveTimeSignatureContext):
        pass

    # Exit a parse tree produced by kernSpineParser#additiveTimeSignature.
    def exitAdditiveTimeSignature(self, ctx:kernSpineParser.AdditiveTimeSignatureContext):
        pass


    # Enter a parse tree produced by kernSpineParser#mixedTimeSignature.
    def enterMixedTimeSignature(self, ctx:kernSpineParser.MixedTimeSignatureContext):
        pass

    # Exit a parse tree produced by kernSpineParser#mixedTimeSignature.
    def exitMixedTimeSignature(self, ctx:kernSpineParser.MixedTimeSignatureContext):
        pass


    # Enter a parse tree produced by kernSpineParser#alternatingTimeSignature.
    def enterAlternatingTimeSignature(self, ctx:kernSpineParser.AlternatingTimeSignatureContext):
        pass

    # Exit a parse tree produced by kernSpineParser#alternatingTimeSignature.
    def exitAlternatingTimeSignature(self, ctx:kernSpineParser.AlternatingTimeSignatureContext):
        pass


    # Enter a parse tree produced by kernSpineParser#alternatingTimeSignatureItem.
    def enterAlternatingTimeSignatureItem(self, ctx:kernSpineParser.AlternatingTimeSignatureItemContext):
        pass

    # Exit a parse tree produced by kernSpineParser#alternatingTimeSignatureItem.
    def exitAlternatingTimeSignatureItem(self, ctx:kernSpineParser.AlternatingTimeSignatureItemContext):
        pass


    # Enter a parse tree produced by kernSpineParser#interchangingTimeSignature.
    def enterInterchangingTimeSignature(self, ctx:kernSpineParser.InterchangingTimeSignatureContext):
        pass

    # Exit a parse tree produced by kernSpineParser#interchangingTimeSignature.
    def exitInterchangingTimeSignature(self, ctx:kernSpineParser.InterchangingTimeSignatureContext):
        pass


    # Enter a parse tree produced by kernSpineParser#meterSymbol.
    def enterMeterSymbol(self, ctx:kernSpineParser.MeterSymbolContext):
        pass

    # Exit a parse tree produced by kernSpineParser#meterSymbol.
    def exitMeterSymbol(self, ctx:kernSpineParser.MeterSymbolContext):
        pass


    # Enter a parse tree produced by kernSpineParser#modernMeterSymbolSign.
    def enterModernMeterSymbolSign(self, ctx:kernSpineParser.ModernMeterSymbolSignContext):
        pass

    # Exit a parse tree produced by kernSpineParser#modernMeterSymbolSign.
    def exitModernMeterSymbolSign(self, ctx:kernSpineParser.ModernMeterSymbolSignContext):
        pass


    # Enter a parse tree produced by kernSpineParser#mensuration.
    def enterMensuration(self, ctx:kernSpineParser.MensurationContext):
        pass

    # Exit a parse tree produced by kernSpineParser#mensuration.
    def exitMensuration(self, ctx:kernSpineParser.MensurationContext):
        pass


    # Enter a parse tree produced by kernSpineParser#metronome.
    def enterMetronome(self, ctx:kernSpineParser.MetronomeContext):
        pass

    # Exit a parse tree produced by kernSpineParser#metronome.
    def exitMetronome(self, ctx:kernSpineParser.MetronomeContext):
        pass


    # Enter a parse tree produced by kernSpineParser#nullInterpretation.
    def enterNullInterpretation(self, ctx:kernSpineParser.NullInterpretationContext):
        pass

    # Exit a parse tree produced by kernSpineParser#nullInterpretation.
    def exitNullInterpretation(self, ctx:kernSpineParser.NullInterpretationContext):
        pass


    # Enter a parse tree produced by kernSpineParser#barline.
    def enterBarline(self, ctx:kernSpineParser.BarlineContext):
        pass

    # Exit a parse tree produced by kernSpineParser#barline.
    def exitBarline(self, ctx:kernSpineParser.BarlineContext):
        pass


    # Enter a parse tree produced by kernSpineParser#barLineType.
    def enterBarLineType(self, ctx:kernSpineParser.BarLineTypeContext):
        pass

    # Exit a parse tree produced by kernSpineParser#barLineType.
    def exitBarLineType(self, ctx:kernSpineParser.BarLineTypeContext):
        pass


    # Enter a parse tree produced by kernSpineParser#restPosition.
    def enterRestPosition(self, ctx:kernSpineParser.RestPositionContext):
        pass

    # Exit a parse tree produced by kernSpineParser#restPosition.
    def exitRestPosition(self, ctx:kernSpineParser.RestPositionContext):
        pass


    # Enter a parse tree produced by kernSpineParser#duration.
    def enterDuration(self, ctx:kernSpineParser.DurationContext):
        pass

    # Exit a parse tree produced by kernSpineParser#duration.
    def exitDuration(self, ctx:kernSpineParser.DurationContext):
        pass


    # Enter a parse tree produced by kernSpineParser#fermata.
    def enterFermata(self, ctx:kernSpineParser.FermataContext):
        pass

    # Exit a parse tree produced by kernSpineParser#fermata.
    def exitFermata(self, ctx:kernSpineParser.FermataContext):
        pass


    # Enter a parse tree produced by kernSpineParser#modernDuration.
    def enterModernDuration(self, ctx:kernSpineParser.ModernDurationContext):
        pass

    # Exit a parse tree produced by kernSpineParser#modernDuration.
    def exitModernDuration(self, ctx:kernSpineParser.ModernDurationContext):
        pass


    # Enter a parse tree produced by kernSpineParser#augmentationDot.
    def enterAugmentationDot(self, ctx:kernSpineParser.AugmentationDotContext):
        pass

    # Exit a parse tree produced by kernSpineParser#augmentationDot.
    def exitAugmentationDot(self, ctx:kernSpineParser.AugmentationDotContext):
        pass


    # Enter a parse tree produced by kernSpineParser#alteration.
    def enterAlteration(self, ctx:kernSpineParser.AlterationContext):
        pass

    # Exit a parse tree produced by kernSpineParser#alteration.
    def exitAlteration(self, ctx:kernSpineParser.AlterationContext):
        pass


    # Enter a parse tree produced by kernSpineParser#staffChange.
    def enterStaffChange(self, ctx:kernSpineParser.StaffChangeContext):
        pass

    # Exit a parse tree produced by kernSpineParser#staffChange.
    def exitStaffChange(self, ctx:kernSpineParser.StaffChangeContext):
        pass


    # Enter a parse tree produced by kernSpineParser#chordSpace.
    def enterChordSpace(self, ctx:kernSpineParser.ChordSpaceContext):
        pass

    # Exit a parse tree produced by kernSpineParser#chordSpace.
    def exitChordSpace(self, ctx:kernSpineParser.ChordSpaceContext):
        pass


    # Enter a parse tree produced by kernSpineParser#graceNote.
    def enterGraceNote(self, ctx:kernSpineParser.GraceNoteContext):
        pass

    # Exit a parse tree produced by kernSpineParser#graceNote.
    def exitGraceNote(self, ctx:kernSpineParser.GraceNoteContext):
        pass


    # Enter a parse tree produced by kernSpineParser#appoggiatura.
    def enterAppoggiatura(self, ctx:kernSpineParser.AppoggiaturaContext):
        pass

    # Exit a parse tree produced by kernSpineParser#appoggiatura.
    def exitAppoggiatura(self, ctx:kernSpineParser.AppoggiaturaContext):
        pass


    # Enter a parse tree produced by kernSpineParser#appoggiaturaMode.
    def enterAppoggiaturaMode(self, ctx:kernSpineParser.AppoggiaturaModeContext):
        pass

    # Exit a parse tree produced by kernSpineParser#appoggiaturaMode.
    def exitAppoggiaturaMode(self, ctx:kernSpineParser.AppoggiaturaModeContext):
        pass


    # Enter a parse tree produced by kernSpineParser#ligatureTie.
    def enterLigatureTie(self, ctx:kernSpineParser.LigatureTieContext):
        pass

    # Exit a parse tree produced by kernSpineParser#ligatureTie.
    def exitLigatureTie(self, ctx:kernSpineParser.LigatureTieContext):
        pass


    # Enter a parse tree produced by kernSpineParser#noteDecoration.
    def enterNoteDecoration(self, ctx:kernSpineParser.NoteDecorationContext):
        pass

    # Exit a parse tree produced by kernSpineParser#noteDecoration.
    def exitNoteDecoration(self, ctx:kernSpineParser.NoteDecorationContext):
        pass


    # Enter a parse tree produced by kernSpineParser#noteDecorationCharX.
    def enterNoteDecorationCharX(self, ctx:kernSpineParser.NoteDecorationCharXContext):
        pass

    # Exit a parse tree produced by kernSpineParser#noteDecorationCharX.
    def exitNoteDecorationCharX(self, ctx:kernSpineParser.NoteDecorationCharXContext):
        pass


    # Enter a parse tree produced by kernSpineParser#phrase.
    def enterPhrase(self, ctx:kernSpineParser.PhraseContext):
        pass

    # Exit a parse tree produced by kernSpineParser#phrase.
    def exitPhrase(self, ctx:kernSpineParser.PhraseContext):
        pass


    # Enter a parse tree produced by kernSpineParser#diatonicPitchAndOctave.
    def enterDiatonicPitchAndOctave(self, ctx:kernSpineParser.DiatonicPitchAndOctaveContext):
        pass

    # Exit a parse tree produced by kernSpineParser#diatonicPitchAndOctave.
    def exitDiatonicPitchAndOctave(self, ctx:kernSpineParser.DiatonicPitchAndOctaveContext):
        pass


    # Enter a parse tree produced by kernSpineParser#trebleNotes.
    def enterTrebleNotes(self, ctx:kernSpineParser.TrebleNotesContext):
        pass

    # Exit a parse tree produced by kernSpineParser#trebleNotes.
    def exitTrebleNotes(self, ctx:kernSpineParser.TrebleNotesContext):
        pass


    # Enter a parse tree produced by kernSpineParser#bassNotes.
    def enterBassNotes(self, ctx:kernSpineParser.BassNotesContext):
        pass

    # Exit a parse tree produced by kernSpineParser#bassNotes.
    def exitBassNotes(self, ctx:kernSpineParser.BassNotesContext):
        pass


    # Enter a parse tree produced by kernSpineParser#accidental.
    def enterAccidental(self, ctx:kernSpineParser.AccidentalContext):
        pass

    # Exit a parse tree produced by kernSpineParser#accidental.
    def exitAccidental(self, ctx:kernSpineParser.AccidentalContext):
        pass


    # Enter a parse tree produced by kernSpineParser#alterationDisplay.
    def enterAlterationDisplay(self, ctx:kernSpineParser.AlterationDisplayContext):
        pass

    # Exit a parse tree produced by kernSpineParser#alterationDisplay.
    def exitAlterationDisplay(self, ctx:kernSpineParser.AlterationDisplayContext):
        pass


    # Enter a parse tree produced by kernSpineParser#turn.
    def enterTurn(self, ctx:kernSpineParser.TurnContext):
        pass

    # Exit a parse tree produced by kernSpineParser#turn.
    def exitTurn(self, ctx:kernSpineParser.TurnContext):
        pass


    # Enter a parse tree produced by kernSpineParser#userAssignable.
    def enterUserAssignable(self, ctx:kernSpineParser.UserAssignableContext):
        pass

    # Exit a parse tree produced by kernSpineParser#userAssignable.
    def exitUserAssignable(self, ctx:kernSpineParser.UserAssignableContext):
        pass


    # Enter a parse tree produced by kernSpineParser#glissando.
    def enterGlissando(self, ctx:kernSpineParser.GlissandoContext):
        pass

    # Exit a parse tree produced by kernSpineParser#glissando.
    def exitGlissando(self, ctx:kernSpineParser.GlissandoContext):
        pass


    # Enter a parse tree produced by kernSpineParser#articulation.
    def enterArticulation(self, ctx:kernSpineParser.ArticulationContext):
        pass

    # Exit a parse tree produced by kernSpineParser#articulation.
    def exitArticulation(self, ctx:kernSpineParser.ArticulationContext):
        pass


    # Enter a parse tree produced by kernSpineParser#accent.
    def enterAccent(self, ctx:kernSpineParser.AccentContext):
        pass

    # Exit a parse tree produced by kernSpineParser#accent.
    def exitAccent(self, ctx:kernSpineParser.AccentContext):
        pass


    # Enter a parse tree produced by kernSpineParser#tenuto.
    def enterTenuto(self, ctx:kernSpineParser.TenutoContext):
        pass

    # Exit a parse tree produced by kernSpineParser#tenuto.
    def exitTenuto(self, ctx:kernSpineParser.TenutoContext):
        pass


    # Enter a parse tree produced by kernSpineParser#staccatissimo.
    def enterStaccatissimo(self, ctx:kernSpineParser.StaccatissimoContext):
        pass

    # Exit a parse tree produced by kernSpineParser#staccatissimo.
    def exitStaccatissimo(self, ctx:kernSpineParser.StaccatissimoContext):
        pass


    # Enter a parse tree produced by kernSpineParser#pizzicato.
    def enterPizzicato(self, ctx:kernSpineParser.PizzicatoContext):
        pass

    # Exit a parse tree produced by kernSpineParser#pizzicato.
    def exitPizzicato(self, ctx:kernSpineParser.PizzicatoContext):
        pass


    # Enter a parse tree produced by kernSpineParser#spiccato.
    def enterSpiccato(self, ctx:kernSpineParser.SpiccatoContext):
        pass

    # Exit a parse tree produced by kernSpineParser#spiccato.
    def exitSpiccato(self, ctx:kernSpineParser.SpiccatoContext):
        pass


    # Enter a parse tree produced by kernSpineParser#staccato.
    def enterStaccato(self, ctx:kernSpineParser.StaccatoContext):
        pass

    # Exit a parse tree produced by kernSpineParser#staccato.
    def exitStaccato(self, ctx:kernSpineParser.StaccatoContext):
        pass


    # Enter a parse tree produced by kernSpineParser#editorialIntervention.
    def enterEditorialIntervention(self, ctx:kernSpineParser.EditorialInterventionContext):
        pass

    # Exit a parse tree produced by kernSpineParser#editorialIntervention.
    def exitEditorialIntervention(self, ctx:kernSpineParser.EditorialInterventionContext):
        pass


    # Enter a parse tree produced by kernSpineParser#slurStart.
    def enterSlurStart(self, ctx:kernSpineParser.SlurStartContext):
        pass

    # Exit a parse tree produced by kernSpineParser#slurStart.
    def exitSlurStart(self, ctx:kernSpineParser.SlurStartContext):
        pass


    # Enter a parse tree produced by kernSpineParser#ligatureTieStart.
    def enterLigatureTieStart(self, ctx:kernSpineParser.LigatureTieStartContext):
        pass

    # Exit a parse tree produced by kernSpineParser#ligatureTieStart.
    def exitLigatureTieStart(self, ctx:kernSpineParser.LigatureTieStartContext):
        pass


    # Enter a parse tree produced by kernSpineParser#tieContinue.
    def enterTieContinue(self, ctx:kernSpineParser.TieContinueContext):
        pass

    # Exit a parse tree produced by kernSpineParser#tieContinue.
    def exitTieContinue(self, ctx:kernSpineParser.TieContinueContext):
        pass


    # Enter a parse tree produced by kernSpineParser#ligatureTieEnd.
    def enterLigatureTieEnd(self, ctx:kernSpineParser.LigatureTieEndContext):
        pass

    # Exit a parse tree produced by kernSpineParser#ligatureTieEnd.
    def exitLigatureTieEnd(self, ctx:kernSpineParser.LigatureTieEndContext):
        pass


    # Enter a parse tree produced by kernSpineParser#slurEnd.
    def enterSlurEnd(self, ctx:kernSpineParser.SlurEndContext):
        pass

    # Exit a parse tree produced by kernSpineParser#slurEnd.
    def exitSlurEnd(self, ctx:kernSpineParser.SlurEndContext):
        pass


    # Enter a parse tree produced by kernSpineParser#barLineCrossedNoteStart.
    def enterBarLineCrossedNoteStart(self, ctx:kernSpineParser.BarLineCrossedNoteStartContext):
        pass

    # Exit a parse tree produced by kernSpineParser#barLineCrossedNoteStart.
    def exitBarLineCrossedNoteStart(self, ctx:kernSpineParser.BarLineCrossedNoteStartContext):
        pass


    # Enter a parse tree produced by kernSpineParser#barLineCrossedNoteEnd.
    def enterBarLineCrossedNoteEnd(self, ctx:kernSpineParser.BarLineCrossedNoteEndContext):
        pass

    # Exit a parse tree produced by kernSpineParser#barLineCrossedNoteEnd.
    def exitBarLineCrossedNoteEnd(self, ctx:kernSpineParser.BarLineCrossedNoteEndContext):
        pass


    # Enter a parse tree produced by kernSpineParser#stem.
    def enterStem(self, ctx:kernSpineParser.StemContext):
        pass

    # Exit a parse tree produced by kernSpineParser#stem.
    def exitStem(self, ctx:kernSpineParser.StemContext):
        pass


    # Enter a parse tree produced by kernSpineParser#beam.
    def enterBeam(self, ctx:kernSpineParser.BeamContext):
        pass

    # Exit a parse tree produced by kernSpineParser#beam.
    def exitBeam(self, ctx:kernSpineParser.BeamContext):
        pass


    # Enter a parse tree produced by kernSpineParser#mordent.
    def enterMordent(self, ctx:kernSpineParser.MordentContext):
        pass

    # Exit a parse tree produced by kernSpineParser#mordent.
    def exitMordent(self, ctx:kernSpineParser.MordentContext):
        pass


    # Enter a parse tree produced by kernSpineParser#trill.
    def enterTrill(self, ctx:kernSpineParser.TrillContext):
        pass

    # Exit a parse tree produced by kernSpineParser#trill.
    def exitTrill(self, ctx:kernSpineParser.TrillContext):
        pass


    # Enter a parse tree produced by kernSpineParser#footnote.
    def enterFootnote(self, ctx:kernSpineParser.FootnoteContext):
        pass

    # Exit a parse tree produced by kernSpineParser#footnote.
    def exitFootnote(self, ctx:kernSpineParser.FootnoteContext):
        pass



del kernSpineParser